@@ -1,5 +1,5 @@
 // Crimson Fields -- a game of tactical warfare
-// Copyright (C) 2000-2006 Jens Granseuer
+// Copyright (C) 2000-2007 Jens Granseuer
 //
 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -53,6 +53,7 @@ enum {
   G_BUTTON_ABORT,
   G_BUTTON_QUIT,
 
+  G_BUTTON_SAVE_AND_SHUTDOWN,
   G_BUTTON_SHUTDOWN,
 
   G_BUTTON_UNIT_INFO,
@@ -78,33 +79,38 @@ public:
                          const char *msg, View *view );
 
   bool Cancelled( void );
+
+private:
+  bool cancelled;
 };
 
 NetworkProgressWindow::NetworkProgressWindow( unsigned short w,
                        unsigned short h, const char *msg, View *view ) :
        ProgressWindow( 0, 0, w, h, 0, 100, msg,
-                       WIN_CENTER|WIN_PROG_ABORT, view ) {}
+                       WIN_CENTER|WIN_PROG_ABORT, view ), cancelled(false) {}
 
 bool NetworkProgressWindow::Cancelled( void ) {
-  if ( Get() < 100 ) Advance( 1 );
-  else Set( 0 );
+  if ( !cancelled ) {
+    if ( Get() < 100 ) Advance( 1 );
+    else Set( 0 );
 
-  bool cancel = ProgressWindow::Cancelled();
+    bool cancel = ProgressWindow::Cancelled();
 
-  if ( cancel ) {
-    // ask for confirmation
-    string buttons;
-    buttons.append( MSG(MSG_B_YES) );
-    buttons += '|';
-    buttons.append( MSG(MSG_B_NO) );
-    DialogWindow *req = new DialogWindow( NULL,
-        MSG(MSG_ASK_ABORT_NETWORK), buttons, 1, 0, view );
-    req->SetButtonID( 1, 0 );
-    cancel = (req->EventLoop() == GUI_CLOSE);
-    view->CloseWindow( req );
+    if ( cancel ) {
+      // ask for confirmation
+      string buttons;
+      buttons.append( MSG(MSG_B_YES) );
+      buttons += '|';
+      buttons.append( MSG(MSG_B_NO) );
+      DialogWindow *req = new DialogWindow( NULL,
+          MSG(MSG_ASK_ABORT_NETWORK), buttons, 1, 0, view );
+      req->SetButtonID( 1, 0 );
+      cancelled = (req->EventLoop() == GUI_CLOSE);
+      view->CloseWindow( req );
+    }
   }
 
-  return cancel;
+  return cancelled;
 }
 #endif
 
@@ -154,7 +160,7 @@ void Game::InitKeys( void ) {
   for ( int i = 0; i <= G_KEY_SWEEP; ++i ) {
     string label( MSG(labels[i]) );
     size_t pos = label.find( '_' );
-    if ( pos != string::npos ) keys[i] = tolower(label[pos+1]);
+    if ( pos != string::npos ) keys[i] = StringUtil::utf8chartoascii(&label[pos+1]);
     else keys[i] = 0;
   }
 }
@@ -359,7 +365,13 @@ int Game::Load( MemBuffer &buffer ) {
     mission->SetLocale( CFOptions.GetLanguage() );
 
     unsigned short flags = mission->GetFlags();
-    if ( !(flags & GI_SAVEFILE) ) {
+    if ( flags & GI_SAVEFILE ) {
+      // update CFOptions
+      if ( flags & GI_CAMPAIGN ) CFOptions.SetCampaign( true );
+      if ( flags & GI_AI ) CFOptions.SetGameType( GTYPE_AI );
+      else if ( flags & GI_PBEM ) CFOptions.SetGameType( GTYPE_PBEM );
+      else if ( flags & GI_NETWORK ) CFOptions.SetGameType( GTYPE_NET_SERVER );
+    } else {
       // if this is a new map (GI_SAVEFILE not set) GI_SKIRMISH,
       // GI_CAMPAIGN, and GI_AI only serve as information
       flags &= ~(GI_AI|GI_CAMPAIGN|GI_SKIRMISH);
@@ -407,6 +419,7 @@ int Game::SwitchMap( const char *sequel ) {
 
   unsigned char p1type = mission->GetPlayer(PLAYER_ONE).Type();
   unsigned char p2type = mission->GetPlayer(PLAYER_TWO).Type();
+  unsigned char handicap = mission->GetHandicap();
 
   // remove current game data
   delete shader;
@@ -422,6 +435,7 @@ int Game::SwitchMap( const char *sequel ) {
   if ( !rc ) {
     mission->GetPlayer(PLAYER_ONE).SetType( p1type );
     mission->GetPlayer(PLAYER_TWO).SetType( p2type );
+    mission->SetHandicap( handicap );
   }
   return rc;
 }
@@ -512,11 +526,13 @@ GUI_Status Game::StartTurn( void ) {
       CheckEvents();
       Execute( remote );
     } else {
-      NoteWindow *ew = new NoteWindow( MSG(MSG_ERROR),
-          MSG(MSG_ERR_NETWORK), WIN_FONT_BIG|WIN_CENTER, view );
-      ew->EventLoop();
-      view->CloseWindow( ew );
-      return GUI_RESTART;
+      if ( pw->Cancelled() ) {
+        // user aborted, so no error
+        return GUI_RESTART;
+      } else {
+        HandleNetworkError();
+        return GUI_OK;
+      }
     }
 
     view->CloseWindow( pw );
@@ -795,13 +811,16 @@ GUI_Status Game::EndTurn( void ) {
         // send data to peer
         History *history = mission->GetHistory();
 
-        if ( !history ) {
-          cerr << "Error: No history available in network game" << endl;
-          return GUI_ERROR;
-        } else {
+        if ( history ) {
           DynBuffer buf;
           history->Save( buf, true );
-          peer->Send( buf );
+          if ( !peer->Send( buf ) ) {
+            HandleNetworkError();
+            return GUI_OK;
+          }
+        } else {
+          cerr << "Error: No history available in network game" << endl;
+          return GUI_ERROR;
         }
       }
 #endif
@@ -1342,33 +1361,26 @@ GUI_Status Game::HandleEvent( const SDL_Event &event ) {
     short key = event.key.keysym.sym & 0x1ff;
     const SDLKey *keymap = CFOptions.GetKeyBindings();
 
-    // map user-defined keys to standard ones
+    // map user- and locale-defined keys to standard ones
     if ((key > SDLK_FIRST) && (key < SDLK_LAST)) {
       // KEYBIND_MINIMIZE is a global binding that is not specific
       // to the Game class, so it isn't handled here
 
-      if (key == keymap[KEYBIND_END_TURN]) key = keys[G_KEY_MAP];
-      else if (key == keymap[KEYBIND_SHOW_MAP]) key = keys[G_KEY_MAP];
-      else if (key == keymap[KEYBIND_UNIT_CONTENT]) key = keys[G_KEY_CONTENT];
-      else if (key == keymap[KEYBIND_UNIT_INFO]) key = keys[G_KEY_INFO];
-      else if (key == keymap[KEYBIND_UNIT_SWEEP]) key = keys[G_KEY_SWEEP];
-      else if (key == keymap[KEYBIND_UNIT_SELECT]) {
-        // select the unit underneath the cursor
-        SelectCommand( mv->Cursor() );
-        key = SDLK_UNKNOWN;
-      } else if (key == keymap[KEYBIND_GAME_MENU]) {
-        GameMenu();
-        key = SDLK_UNKNOWN;
-      } else if (key == keymap[KEYBIND_UNIT_MENU]) {
-        Unit *u = map.GetUnit( mv->Cursor() );
-        if ( u ) UnitMenu( u );
-        key = SDLK_UNKNOWN;
-      } else if (key == keymap[KEYBIND_UNIT_UNDO]) {
-        if ( undo.GetUnit() != NULL) Undo();
-        key = SDLK_UNKNOWN;
-      } else if (key == keymap[KEYBIND_UNIT_NEXT]) {
-        SelectNextUnit();
-        key = SDLK_UNKNOWN;
+      // user-defined keys first
+      for ( int i = 0; i < KEYBIND_COUNT; ++i ) {
+        if ( key == keymap[i] ) {
+          key = -i;
+          break;
+        }
+      }
+
+      // now check locale-defined keys
+      if ( key >= 0 ) {
+        if ( key == keys[G_KEY_INFO] ) key = -KEYBIND_UNIT_INFO;
+        else if ( key == keys[G_KEY_CONTENT] ) key = -KEYBIND_UNIT_CONTENT;
+        else if ( key == keys[G_KEY_SWEEP] ) key = -KEYBIND_UNIT_SWEEP;
+        else if ( key == keys[G_KEY_END_TURN] ) key = -KEYBIND_END_TURN;
+        else if ( key == keys[G_KEY_MAP] ) key = -KEYBIND_SHOW_MAP;
       }
     }
 
@@ -1391,34 +1403,58 @@ GUI_Status Game::HandleEvent( const SDL_Event &event ) {
       }
       GameMenu();
       break;
-    default:
-      // these shortcuts change with the locale
 
-      if ( key == keys[G_KEY_INFO] ) {
-        Unit *u = map.GetUnit( mv->Cursor() );
-        if ( u ) UnitInfo( u );
-      } else if ( key == keys[G_KEY_MAP] ) {
-        new TacticalWindow( mv, *mission, view );
-      } else if ( key == keys[G_KEY_CONTENT] ) {
-        MapObject *mo = map.GetMapObject( mv->Cursor() );
-        if ( mo ) {
-          if ( mo->IsShop() )
-            ContainerContent( static_cast<UnitContainer *>(
-                              static_cast<Building *>(mo)) );
-          else if ( static_cast<Unit *>(mo)->IsTransport() )
-            ContainerContent( static_cast<UnitContainer *>(
-                              static_cast<Transport *>(mo)) );
-        }
-      } else if ( key == keys[G_KEY_END_TURN] ) {
-        rc = EndTurn();
-      } else if ( key == keys[G_KEY_SWEEP] ) {
-        if ( unit && MinesweeperTargets( unit ) )
-          EnterSpecialMode( MODE_SWEEP );
-      } else if ( key == keys[G_KEY_QUIT] ) {
-        Quit();
-      }
+    // the keys for the following commands can be modified by
+    // user and/or locale, so we use fixed (negative so we don't
+    // collide with regular keysyms) values which get mapped
+    // before the switch statement
+    case -KEYBIND_END_TURN:
+      rc = EndTurn();
+      break;
+    case -KEYBIND_SHOW_MAP:
+      new TacticalWindow( mv, *mission, view );
+      break;
+    case -KEYBIND_GAME_MENU:
+      GameMenu();
+      break;
+    case -KEYBIND_UNIT_MENU: {
+      Unit *u = map.GetUnit( mv->Cursor() );
+      if ( u ) UnitMenu( u ); }
+      break;
+    case -KEYBIND_UNIT_CONTENT: {
+      MapObject *mo = map.GetMapObject( mv->Cursor() );
+      if ( mo ) {
+        if ( mo->IsShop() )
+          ContainerContent( static_cast<UnitContainer *>(
+                            static_cast<Building *>(mo)) );
+        else if ( static_cast<Unit *>(mo)->IsTransport() )
+          ContainerContent( static_cast<UnitContainer *>(
+                            static_cast<Transport *>(mo)) );
+      }}
+      break;
+    case -KEYBIND_UNIT_INFO: {
+      Unit *u = map.GetUnit( mv->Cursor() );
+      if ( u ) UnitInfo( u ); }
+      break;
+    case -KEYBIND_UNIT_NEXT:
+      SelectNextUnit();
+      break;
+    case -KEYBIND_UNIT_SELECT:
+      // select the unit underneath the cursor
+      SelectCommand( mv->Cursor() );
+      break;
+    case -KEYBIND_UNIT_UNDO:
+      if ( undo.GetUnit() != NULL) Undo();
+      break;
+    case -KEYBIND_UNIT_SWEEP:
+      if ( unit && MinesweeperTargets( unit ) )
+        EnterSpecialMode( MODE_SWEEP );
+      break;
+    default:
+      if ( key == keys[G_KEY_QUIT] ) Quit();
       break;
     }
+
   } else if ( event.type == SDL_MOUSEBUTTONDOWN ) {
     Point pos;
     if ( event.button.button == SDL_BUTTON_LEFT ) {
@@ -1863,8 +1899,8 @@ void Game::GameMenu( void ) {
   menu->AddItem( 1, G_BUTTON_KEYBOARD_OPTIONS, 0, MSG(MSG_B_OPT_KEYBOARD) );
   menu->AddBar( 0 );
   menu->AddItem( 0, G_BUTTON_SAVE,
-        (mission->GetFlags() & GI_PBEM || CFOptions.GetGameType() == GTYPE_NET_CLIENT) ?
-        WIDGET_DISABLED : 0, MSG(MSG_B_SAVE_GAME) );
+        mission->GetFlags() & GI_PBEM ? WIDGET_DISABLED : 0,
+        MSG(MSG_B_SAVE_GAME) );
 
   if (view->IsFullScreen())
     menu->AddItem( 0, G_BUTTON_MINIMIZE_WINDOW, 0, MSG(MSG_OPT_KEY_MINIMIZE) );
@@ -1905,6 +1941,28 @@ void Game::UnitMenu( Unit *u ) {
   } else UnitInfo( u );
 }
 
+////////////////////////////////////////////////////////////////////////
+// NAME       : Game::HandleNetworkError
+// DESCRIPTION: Call when a network error has been detected. Ask user
+//              whether current game should be saved, and shutdown
+//              afterwards.
+// PARAMETERS : -
+// RETURNS    : -
+////////////////////////////////////////////////////////////////////////
+
+void Game::HandleNetworkError( void ) {
+  string choices( MSG(MSG_B_SAVE) );
+  choices += '|';
+  choices += MSG(MSG_B_CANCEL);
+
+  DialogWindow *dw = new DialogWindow( MSG(MSG_ERROR),
+      MSG(MSG_ERR_NETWORK), choices, 0, WIN_FONT_BIG|WIN_CENTER,
+      view );
+  dw->SetButtonHook( this );
+  dw->SetButtonID( 0, G_BUTTON_SAVE_AND_SHUTDOWN );
+  dw->SetButtonID( 1, G_BUTTON_SHUTDOWN );
+}
+
 ////////////////////////////////////////////////////////////////////////
 // NAME       : Game::Shutdown
 // DESCRIPTION: This method should be called when the current game ends,
@@ -1946,10 +2004,6 @@ GUI_Status Game::WidgetActivated( Widget *button, Window *win ) {
     view->CloseWindow( win );
     ShowBriefing();
     break;
-  case G_BUTTON_SAVE:
-    view->CloseWindow( win );
-    Save( NULL );
-    break;
   case G_BUTTON_LEV_INFO:
     view->CloseWindow( win );
     ShowLevelInfo();
@@ -1997,6 +2051,13 @@ GUI_Status Game::WidgetActivated( Widget *button, Window *win ) {
     view->CloseWindow( win );
     Quit();
     break;
+  case G_BUTTON_SAVE:
+  case G_BUTTON_SAVE_AND_SHUTDOWN:
+    view->CloseWindow( win );
+    Save( NULL );
+    if ( button->ID() == G_BUTTON_SAVE )
+      break;
+    // otherwise fall through
   case G_BUTTON_SHUTDOWN:
     Shutdown();
     rc = GUI_RESTART;
